function [f, df, filledY] = neglogprob_of_everything(realparams, params, Y, w, factor_idx, ... 
    segments, MGPopts, Kn, invKn, ldKn, nmodeKn, nmodeNn, Kpn, invKpn, ldKpn, WVariance)
% Compute the objective function (negative log likelihood) for jointly 
% optimizing input and hyperparameters in L-BFGS
% 
% params - parameters being optimized 
%          e.g. for 2 factors, 
%          params = [Us{1}(:)' log(Hps{1}) log(Phps{1}) 
%                    Us{2}(:)' log(Hps{2}) log(Phps{2})
%                    log(theta) log(beta)];
%          using notations from computeNModeKernel.m, plus 
%          Phps - cell array of hyperparameters for priors on factor inputs
%                 e.g., GPDM or Gaussian
%          WVariance - variance of the Gaussian prior on w, 0 if no prior
% Inside struct MGPopts{n} (for the n-th factor): 
% qs - latent dimensions fro each factor
% Ns - number of data points for each factor 
% Nhps - number of hyperparameters for each factor
% Nphps - number of hyperparameters for input priors on each factor
% Y - N x D data matrix
% factor_idx - see idx in computeNModeKernel.m
% w - 1 x D scales for each output dimension
% meanS, covS - mean and covariances if Gaussian input priors are used
% segments - cell array of where each time series segment starts for each
%            factor, empty for independent data
% priorModel - cell array GPDM model type for each factor, only used if 
%              Phps is not empty
% G - control points for b-spline (not in use right now)
% periodic - extra constraints on the latent space for each factor
%            0 - no extra constraint Us{i}(:)' is used in params as above
%            1 - circular latent space, where each data point is assigned 
%                an angle, t{i} a 1 x Ns(i) vector is used in params
%            2 - circular latent space, where each motion segment is 
%                assigned an angle and a stepsize, t{i} is 1 x 2*number of
%                segments
% BalanceTerm - B-GPDM factor if GPDM is used
%
% Global variables: 
% OPT_MASK - vector of indices in params to keep fixed during the
%            optimization
% 
% L_BFGS - whether the L-BFGS optimizer is used
%

global OPT_MASK L_BFGS;  

noscale = 0; 

missingData = isnan(Y);
HAVE_DYNAMIC_KERNEL = 0; 
if (nargin < 8) 
    HAVE_OLD_KERNEL = 0; 
else
    HAVE_OLD_KERNEL = 1; 
    if (nargin > 11) 
        HAVE_DYNAMIC_KERNEL = 1; 
    end
end

if (nargin < 16) 
    WVariance = 0; 
end


NO_HYPERPRIOR = 1; 
if (L_BFGS == 1) 
    realparams = realparams';
end

optIdx = setdiff(1:length(params),OPT_MASK); 

params(optIdx) = realparams; 

nfactor = size(factor_idx, 2); 

qs = zeros(1,nfactor); Ns = zeros(1,nfactor); Nhps = zeros(1,nfactor); Nphps = zeros(1,nfactor);
priorModel = cell(nfactor); periodic = zeros(1,nfactor);  
for n = 1:nfactor
    qs(n) = MGPopts{n}.qs; 
    Ns(n) = MGPopts{n}.Ns; 
    Nhps(n) = MGPopts{n}.Nhps; 
    Nphps(n) = MGPopts{n}.Nphps; 
    priorModel{n} = MGPopts{n}.priorModel; 
    periodic(n) = MGPopts{n}.periodic; 
    meanS{n} = MGPopts{n}.meanS; 
    covS{n} = MGPopts{n}.covS; 
    G{n} = MGPopts{n}.G;
    lambda(n) = MGPopts{n}.BalanceTerm; 
end

N = size(Y, 1);
D = size(Y, 2);
if (sum(w) == D)
    noscale = 1; 
end
currentIdx = 0; 
for i = 1:nfactor
    if (~isempty(G{i}))
        dUs_dt{i} = zeros(Ns(i),qs(i));
        dUs_dG{i} = zeros(Ns(i), size(G,1)); 
        G{i} = reshape(params([1:size(G{i},1)*qs(i)] + currentIdx), size(G{i},1), qs(i));
        currentIdx = currentIdx + size(G{i},1)*qs(i);
        t{i} = params([1:Ns(i)] + currentIdx);
        currentIdx = currentIdx + Ns(i);
        for j = 1:Ns(i)
            [Us{i}(j,:) dUs_dt{i}(j,:) tmp] = b_spline(G{i}, t{i}(j));
            for k = 1:size(G{i},1)
                dUs_dG{i}(j,k) = tmp(k,1); 
            end   
        end
    else
        if (periodic(i) == 1) 
            t{i} = params([1:Ns(i)] + currentIdx); 
            currentIdx = currentIdx + Ns(i);
            for j = 1:Ns(i)
                Us{i}(j,:) = [cos(t{i}(j)) sin(t{i}(j))]; 
                dUs_dt{i}(j,:) = [-sin(t{i}(j)) cos(t{i}(j))]; 
            end
        elseif (periodic(i) == 2)
            t{i} = params([1:2*length(segments{i})] + currentIdx); 
            currentIdx = currentIdx + 2*length(segments{i});
            leftovers = reshape(params([1:Ns(1)*(qs(1)-2)] + currentIdx), Ns(1), qs(1)-2);
            currentIdx = currentIdx + Ns(1)*(qs(1)-2); 
            for n = 1:length(segments{i})-1
                dUs_dt{i}{n} = zeros(Ns(i),2);
                dUs_dt{i}{length(segments{i})+n} = zeros(Ns(i),2);
                theta_0 = t{i}(n); 
                delta = t{i}(length(segments{i})+n); 
                for j = segments{i}(n):segments{i}(n+1)-1
                    j_1 = j - segments{i}(n);
                    Us{i}(j,:) = [cos(theta_0 + (j_1)*delta) sin(theta_0 + (j_1)*delta)];

                    dUs_dt{i}{n}(j,:) = ...
                        [-sin(theta_0 + (j_1)*delta) cos(theta_0 + (j_1)*delta)];
                    dUs_dt{i}{length(segments{i})+n}(j,:) = ...
                        [-sin(theta_0 + (j_1)*delta)*(j_1) cos(theta_0 + (j_1)*delta)*(j_1)];
                end
             
                if (~isempty(leftovers))
                    dUs_dt{i}{n} = [ dUs_dt{i}{n} zeros(Ns(i),qs(i)-2)];
                    dUs_dt{i}{length(segments{i})+n} = [dUs_dt{i}{length(segments{i})+n} zeros(Ns(i),qs(i)-2)];
                end
            end
            n = length(segments{i});
            dUs_dt{i}{n} = zeros(Ns(i),2);
            dUs_dt{i}{length(segments{i})+n} = zeros(Ns(i),2);
            theta_0 = t{i}(n);
            delta = t{i}(length(segments{i})+n);
            for j = segments{i}(n):Ns(i)
                j_1 = j - segments{i}(n);
                Us{i}(j,:) = [cos(theta_0 + (j_1)*delta) sin(theta_0 + (j_1)*delta)];
                dUs_dt{i}{n}(j,:) = ...
                    [-sin(theta_0 + (j_1)*delta) cos(theta_0 + (j_1)*delta)];
                dUs_dt{i}{length(segments{i})+n}(j,:) = ...
                    [-sin(theta_0 + (j_1)*delta)*(j_1) cos(theta_0 + (j_1)*delta)*(j_1)];
            end
            if (~isempty(leftovers))
                                    dUs_dt{i}{n} = [ dUs_dt{i}{n} zeros(Ns(i),qs(i)-2)];
                    dUs_dt{i}{length(segments{i})+n} = [dUs_dt{i}{length(segments{i})+n} zeros(Ns(i),qs(i)-2)];
                      end
            Us{i} = [Us{i} leftovers];

        else      
            Us{i} = reshape(params([1:Ns(i)*qs(i)] + currentIdx), Ns(i), qs(i));
            currentIdx = currentIdx + Ns(i)*qs(i);
        end
    end
    Hps{i} = exp(params([1:Nhps(i)] + currentIdx));
    currentIdx = currentIdx + Nhps(i);
    Phps{i} = exp(params([1:Nphps(i)] + currentIdx));
    currentIdx = currentIdx + Nphps(i);
end

if (size(missingData, 2) == 1)
    completeFrames = find(missingData == 0);
    missingFrames = find(missingData > 0);
else
    completeFrames = find(sum(missingData') == 0);
    missingFrames = find(sum(missingData') > 0);
end


globHps = exp(params(end-1:end));

if HAVE_OLD_KERNEL == 1
    if nargout > 1
        [K, invK, ldK, junk1, junk2, dC_dtheta, dC_dU, dC_dgamma] = ...
            computeNModeKernel(Us, Hps, factor_idx(completeFrames,:), globHps(1), globHps(2), Kn, invKn, ldKn, nmodeKn, nmodeNn);
    else
        [K, invK, ldK] = ...
            computeNModeKernel(Us, Hps, factor_idx(completeFrames,:), globHps(1), globHps(2), Kn, invKn, ldKn, nmodeKn, nmodeNn);
    end
else
    if nargout > 1
        [K, invK, ldK, junk1, junk2, dC_dtheta, dC_dU, dC_dgamma] = ...
            computeNModeKernel(Us, Hps, factor_idx(completeFrames,:), globHps(1), globHps(2));
    else
        [K, invK, ldK] = ...
            computeNModeKernel(Us, Hps, factor_idx(completeFrames,:), globHps(1), globHps(2));
    end
end

filledY = Y;
if (~isempty(missingFrames))
    % If there are missing inputs, fill it in using GP mean prediction,
    % this is equivalent to maximizing the likelihood of y given the latent variables. 
    M = length(missingFrames);
    N = length(completeFrames);
    D = size(Y,2);
    factor_map = factor_idx(:,1);
    factor_map(completeFrames) = 1:N;
    factor_map(missingFrames) = 1:M;

    Ka = ones(N, M);

    for n = 1:nfactor
        U = Us{n}(factor_idx(completeFrames,n),:);
        V = Us{n}(factor_idx(missingFrames,n),:);
        gamma = Hps{n};

        if isempty(gamma)
            Kna = U*V';
        else
            N2 = dist2(U, V);
            Kna = exp(-N2*gamma/2);
        end
        Ka = Ka.*Kna(factor_map(factor_idx(completeFrames,n)), factor_map(factor_idx(missingFrames,n)));
    end

    A = globHps(1)*Ka'*invK;
    B = A*Y(completeFrames,:);

    [rIdx cIdx] = find(missingData(missingFrames,:) > 0);
    Tmp = Y(missingFrames,:);
    Tmp(rIdx, cIdx) = B(rIdx, cIdx);
    Y(missingFrames,:) = Tmp;
    filledY = Y;
    if HAVE_OLD_KERNEL == 1
        if nargout > 1
            [K, invK, ldK, junk1, junk2, dC_dtheta, dC_dU, dC_dgamma] = ...
                computeNModeKernel(Us, Hps, factor_idx, globHps(1), globHps(2), Kn, invKn, ldKn, nmodeKn, nmodeNn);
        else
            [K, invK, ldK] = ...
                computeNModeKernel(Us, Hps, factor_idx, globHps(1), globHps(2), Kn, invKn, ldKn, nmodeKn, nmodeNn);
        end
    else
        if nargout > 1
            [K, invK, ldK, junk1, junk2, dC_dtheta, dC_dU, dC_dgamma] = ...
                computeNModeKernel(Us, Hps, factor_idx, globHps(1), globHps(2));
        else
            [K, invK, ldK] = ...
                computeNModeKernel(Us, Hps, factor_idx, globHps(1), globHps(2));
        end
    end
    N = N+M;
end

L = -D*N/2*log(2*pi)-D/2*ldK;

if (noscale == 0)
    for d= 1:D
        Yscaled(:,d) = w(d)*Y(:,d);
    end
    L = L -0.5*trace(invK*Yscaled*Yscaled');
    L = L + N*sum(log(w));

    if (WVariance > 0)
        L = L + D*log(2) - D/2*log(2*pi*WVariance) - 0.5/WVariance*sum(w.*w) ;
    end
else
    L = L -0.5*trace(invK*Y*Y');
end

if (NO_HYPERPRIOR == 0)
    L = L - sum(log([globHps(1) globHps(2)]));

    for i = 1:nfactor
        if (~isempty(Hps{i}))
            L = L - log(Hps{i});
        end
    end
end

Lp = 0; 
cacheKp = cell(1, nfactor); 
for i = 1:nfactor
    if (isempty(Phps{i}))
        if (isempty(covS{i}))
            if (~isempty(Hps{i}))
                %Lp = Lp - Ns(i)*qs(i)/2*log(2*pi) - 0.5*sum(sum(Us{i}.*Us{i}));
            end
        else
            Lp = Lp - Ns(i)*qs(i)/2*log(2*pi) - Ns(i)/2*logdet(covS{i}); 
            icovS = pdinv(covS{i}); 
            for n = 1:Ns(i)
                Lp = Lp - 0.5*(Us{i}(n, :) - meanS{i})*icovS*(Us{i}(n, :) - meanS{i})'; 
            end
        end
    else
        [Xin Xout] = priorIO(Us{i}, segments{i}, priorModel{i});
        Np = size(Xin, 1);
        qp = size(Xin, 2); 
        if (priorModel{i}(3) == -1)
            Lp = Lp - Ns(i)*qs(i)/2*log(2*pi) - 0.5*sum(sum(Us{i}.*Us{i}));
        else
            if (HAVE_DYNAMIC_KERNEL == 1)
                if nargout > 1
                    [Kp, invKp, cacheKp{i}, ldKp] = computePriorKernel(Xin, Phps{i}, priorModel{i}(3), Kpn{i}, invKpn{i}, ldKpn{i});
                else
                    [Kp, invKp, junk, ldKp] = computePriorKernel(Xin, Phps{i}, priorModel{i}(3), Kpn{i}, invKpn{i}, ldKpn{i});
                end
            else
                if nargout > 1
                    [Kp, invKp, cacheKp{i}, ldKp] = computePriorKernel(Xin, Phps{i}, priorModel{i}(3));
                else
                    [Kp, invKp, junk, ldKp] = computePriorKernel(Xin, Phps{i}, priorModel{i}(3));
                end
            end

            
            Lp = Lp - qs(i)*Np/2*log(2*pi)-qs(i)/2*ldKp;
 

            for d= 1:qs(i)
                Lp = Lp - 0.5*Xout(:, d)'*invKp*Xout(:, d);
            end

            if (priorModel{i}(1) == 0 || priorModel{i}(1) == 1)
                XDiff = Us{i}(segments{i}+1,:) - Us{i}(segments{i},:);
                Lp = Lp - size(segments{i},2)*qs(i)/2*log(2*pi) - 0.5*sum(sum(XDiff.*XDiff));
            end

            Lp = Lp - size(segments{i},2)*qs(i)/2*log(2*pi) - 0.5*sum(sum(Us{i}(segments{i},:).*Us{i}(segments{i},:)));
            Lp = lambda(i)*Lp; 
            if (NO_HYPERPRIOR == 0)
                Lp = Lp - sum(log(Phps{i}));
            end
        end
    end
end

L = L + Lp;

L = -L;
f = L; 

if nargout > 1
    if (noscale == 0)
        for d=1:D
            Yscaled(:,d) = w(d)*Y(:,d);
        end

        tmp = invK*Yscaled;
    else
        tmp = invK*Y;
    end
    dL_dK = -D/2*invK + .5*tmp*tmp';

    g = [];

    for i = 1:nfactor
        dL_dUk = zeros(N, qs(i));
        dL_dU = zeros(Ns(i), qs(i));
        for d = 1:qs(i)
            dL_dUk(:,d) = 2*sum(dL_dK.*dC_dU{i}{d}, 2) - diag(dL_dK).*diag(dC_dU{i}{d});
        end
        for d = 1:Ns(i)
            dL_dU(d,:) = sum(dL_dUk(find(factor_idx(:,i) == d),:),1);
        end

        gParamp = zeros(1,Nphps(i));
        if (isempty(Phps{i}))
            if (isempty(covS{i}))
                if (~isempty(Hps{i}))
                    %dL_dU = dL_dU - Us{i};
                end
            else
                icovS = pdinv(covS{i});             
                for n = 1:Ns(i)
                   dL_dU(n,:) = dL_dU(n,:) - 0.5*((icovS'+icovS)*(Us{i}(n, :) - meanS{i})')'; 
                end
            end
        else
            if (priorModel{i}(3) == -1)
                dLp_dU = -Us{i};
            else
                [Xin Xout] = priorIO(Us{i}, segments{i}, priorModel{i});
                thetap = Phps{i}; 

                Np = size(Xin, 1);

                tmp = invKp*Xout; 
                dLp_dK = -qs(i)/2*invKp + 0.5*tmp*tmp';
                dLp_dK = lambda(i)*dLp_dK; 

                dk = zeros(1, Nphps(i));
                if (priorModel{i}(3) == 0)
                    [dK{1}, dK{2}, dK{3}] = kernel2DiffParams(Xin, thetap);
                elseif (priorModel{i}(3) == 1)
                    [dK{1}, dK{2}] = kernelDiffParams(Xin, Xin, thetap, cacheKp{i});
                elseif (priorModel{i}(3) == 2)
                    [dK{1}] = lin_kernelDiffParams(Xin, thetap);
                elseif (priorModel{i}(3) == 3)
                    [dK{1} dK{2}] = lin_kernel2DiffParams(Xin, thetap(1:2));
                    [dK{3}, dK{4}, dK{5}] = kernel2DiffParams(Xin, thetap(3:5));
                elseif (priorModel{i}(3) == 4)
                    [dK{1} dK{2}] = lin_kernel2DiffParams(Xin, thetap);
                elseif (priorModel{i}(3) == 5)
                    [dK{1}] = lin_kernelDiffParams(Xin, thetap(1));
                    [dK{2}, dK{3}] = kernelDiffParams(Xin, Xin, thetap(2:3), cacheKp{i});
                end

                for j = 1:(Nphps(i)-1)
                    dk(j) = sum(sum(dLp_dK.*dK{j}));
                end
                dk(Nphps(i)) = -sum(diag(dLp_dK)/thetap(end).^2);
                if (NO_HYPERPRIOR == 0)
                    gParamp = dk.*thetap - 1;
                else
                    gParamp = dk.*thetap; 
                end

                dLp_dxin = zeros(Np, qp);

                for d = 1:qp
                    if (priorModel{i}(3) == 0)
                        Kpart = kernel2DiffX(Xin, thetap, d);
                    elseif (priorModel{i}(3) == 1)
                        Kpart = kernelDiffX(Xin, thetap, d, cacheKp{i});
                    elseif (priorModel{i}(3) == 2)
                        Kpart = lin_kernelDiffX(Xin, thetap, d);
                    elseif (priorModel{i}(3) == 3)
                        Kpart = lin_kernel2DiffX(Xin, thetap(1:2), d) + ...
                            kernel2DiffX(Xin, thetap(3:5), d);
                    elseif (priorModel{i}(3) == 4)
                        Kpart = lin_kernel2DiffX(Xin, thetap, d);
                    elseif (priorModel{i}(3) == 5)
                        Kpart = lin_kernelDiffX(Xin, thetap(1), d) + ...
                            kernelDiffX(Xin, thetap(2:3), d, cacheKp{i});
                    end
                    dLp_dxin(:, d) = ...
                        2*sum(dLp_dK.*Kpart, 2) - diag(dLp_dK).*diag(Kpart);
                end

                dLp_dU = priorDiffX(dLp_dxin, -lambda(i)*invKp*Xout, Ns(i), qs(i), segments{i}, priorModel{i});

                if (priorModel{i}(1) == 0 || priorModel{i}(1) == 1)
                    dLp_dU(segments{i}+1,:) = dLp_dU(segments{i}+1,:) - lambda(i)*(Us{i}(segments{i}+1,:) - Us{i}(segments{i},:));
                    dLp_dU(segments{i},:) = dLp_dU(segments{i},:) - lambda(i)*Us{i}(segments{i},:) + lambda(i)*(Us{i}(segments{i}+1,:) - Us{i}(segments{i},:));
                elseif (priorModel{i}(1) == 2 || priorModel{i}(1) == 3)
                    dLp_dU(segments{i},:) = dLp_dU(segments{i},:) - lambda(i)*Us{i}(segments{i},:);
                end
            end
            dL_dU = dL_dU + dLp_dU;
        end

        if (~isempty(G{i}))
            for j = 1:Ns(i)
                tmp = zeros(size(dUs_dt{i}));
                tmp(j,:) = dUs_dt{i}(j,:);
                dL_dt(j) = trace(dL_dU'*tmp);
            end

            for j = 1:size(G{i},1)
                for k = 1:qs(i)
                    tmp = zeros(Ns(i),qs(i));
                    tmp(:,k) = dUs_dG{i}(:,j);
                    dL_dG(j,k) = trace(dL_dU'*tmp);
                end
            end
            g = [g dL_dG(:)' dL_dt];
        else
            if (periodic(i) == 1)
                for j = 1:Ns(i)
                    tmp = zeros(size(dUs_dt{i}));
                    tmp(j,:) = dUs_dt{i}(j,:);
                    dL_dt(j) = trace(dL_dU'*tmp);
                end
                g = [g dL_dt(:)']; 
            elseif (periodic(i) == 2) 
                for n = 1:length(segments{i})
                    dL_dt(n) = trace(dL_dU'*dUs_dt{i}{n}); 
                    dL_dt(length(segments{i})+n) = trace(dL_dU'*dUs_dt{i}{length(segments{i})+n}); 
                end
                if (~isempty(leftovers))
                    dL_dleftovers = dL_dU(:,3:end); 
                    g = [g dL_dt(:)' dL_dleftovers(:)'];
                else
                    g = [g dL_dt(:)'];
                end
            else
                g = [g dL_dU(:)'];
            end
        end

        if (~isempty(dC_dgamma{i}))
            if (NO_HYPERPRIOR == 0) 
                g = [g sum(sum(dL_dK.*dC_dgamma{i}))*Hps{i} - 1];
            else
                g = [g sum(sum(dL_dK.*dC_dgamma{i}))*Hps{i}];
            end
        end

        if (~isempty(Phps{i}))
            g = [g gParamp]; 
        end
    end

    if (NO_HYPERPRIOR == 0)
        g = [g, sum(sum(dL_dK.*dC_dtheta))*globHps(1)-1, -sum(diag(dL_dK)/globHps(2).^2)*globHps(2)-1];
    else
        g = [g, sum(sum(dL_dK.*dC_dtheta))*globHps(1), -sum(diag(dL_dK)/globHps(2).^2)*globHps(2)];
    end
    df = -g;
    df = df(optIdx);   
end
